<?php

namespace Drupal\tmgmt_memsource\Plugin\tmgmt\Translator;

use Drupal\Core\Plugin\ContainerFactoryPluginInterface;
use Drupal\tmgmt\ContinuousTranslatorInterface;
use Drupal\tmgmt\Entity\RemoteMapping;
use Drupal\tmgmt\JobInterface;
use Drupal\tmgmt\JobItemInterface;
use Drupal\tmgmt\TMGMTException;
use Drupal\tmgmt\Translator\AvailableResult;
use Drupal\tmgmt\TranslatorInterface;
use Drupal\tmgmt\TranslatorPluginBase;
use GuzzleHttp\ClientInterface;
use GuzzleHttp\Exception\RequestException;
use Symfony\Component\DependencyInjection\ContainerInterface;

/**
 * Memsource translation plugin controller.
 *
 * @TranslatorPlugin(
 *   id = "memsource",
 *   label = @Translation("Memsource"),
 *   description = @Translation("Memsource translator service."),
 *   ui = "Drupal\tmgmt_memsource\MemsourceTranslatorUi",
 * )
 */
class MemsourceTranslator extends TranslatorPluginBase implements ContainerFactoryPluginInterface, ContinuousTranslatorInterface {

  /**
   * The translator.
   *
   * @var \Drupal\tmgmt\TranslatorInterface
   */
  private $translator;

  /**
   * Guzzle HTTP client.
   *
   * @var \GuzzleHttp\ClientInterface
   */
  protected $client;

  /**
   * Constructs a MemsourceTranslator object.
   *
   * @param \GuzzleHttp\ClientInterface $client
   *   The Guzzle HTTP client.
   * @param array $configuration
   *   A configuration array containing information about the plugin instance.
   * @param string $plugin_id
   *   The plugin_id for the plugin instance.
   * @param array $plugin_definition
   *   The plugin implementation definition.
   */
  public function __construct(ClientInterface $client, array $configuration, $plugin_id, array $plugin_definition) {
    parent::__construct($configuration, $plugin_id, $plugin_definition);
    $this->client = $client;
  }

  /**
   * {@inheritdoc}
   */
  public static function create(ContainerInterface $container, array $configuration, $plugin_id, $plugin_definition) {
    /** @var \GuzzleHttp\ClientInterface $client */
    $client = $container->get('http_client');
    return new static(
      $client,
      $configuration,
      $plugin_id,
      $plugin_definition
    );
  }

  /**
   * Sets a Translator.
   *
   * @param \Drupal\tmgmt\TranslatorInterface $translator
   *   The translator to set.
   */
  public function setTranslator(TranslatorInterface $translator) {
    $this->translator = $translator;
  }

  /**
   * {@inheritdoc}
   */
  public function getSupportedRemoteLanguages(TranslatorInterface $translator) {
    $supported_remote_languages = [];
    $this->setTranslator($translator);
    try {
      $supported_languages = $this->sendApiRequest('/api2/v1/languages');
      foreach ($supported_languages['languages'] as $language) {
        $supported_remote_languages[$language['code']] = $language['name'];
      }
    }
    catch (\Exception $e) {
      // Ignore exception, nothing we can do.
    }
    asort($supported_remote_languages);
    return $supported_remote_languages;
  }

  /**
   * {@inheritdoc}
   */
  public function checkAvailable(TranslatorInterface $translator) {
    if ($this->getToken()) {
      return AvailableResult::yes();
    }
    return AvailableResult::no(t('@translator is not available. Make sure it is properly <a href=:configured>configured</a>.', [
      '@translator' => $translator->label(),
      ':configured' => $translator->toUrl()->toString(),
    ]));
  }

  /**
   * {@inheritdoc}
   */
  public function requestTranslation(JobInterface $job) {
    $job = $this->requestJobItemsTranslation($job->getItems());
    if (!$job->isRejected()) {
      $job->submitted();
    }
  }

  /**
   * {@inheritdoc}
   */
  public function requestJobItemsTranslation(array $job_items) {
    /** @var \Drupal\tmgmt\Entity\Job $job */
    $job = reset($job_items)->getJob();
    $this->setTranslator($job->getTranslator());
    $due_date = $job->getSetting('due_date');

    try {
      $project_id = $this->newTranslationProject($job);
      $job->addMessage('Created a new project in Memsource with the id: @id', ['@id' => $project_id], 'debug');

      /** @var \Drupal\tmgmt\Entity\JobItem $job_item */
      foreach ($job_items as $job_item) {
        $job_part_id = $this->sendFiles($job_item, $project_id, $due_date);

        /** @var \Drupal\tmgmt\Entity\RemoteMapping $remote_mapping */
        $remote_mapping = RemoteMapping::create([
          'tjid' => $job->id(),
          'tjiid' => $job_item->id(),
          'remote_identifier_1' => 'tmgmt_memsource',
          'remote_identifier_2' => $project_id,
          'remote_identifier_3' => $job_part_id,
          'remote_data' => [
            'FileStateVersion' => 1,
            'TmsState' => 'TranslatableSource',
            'RequiredBy' => $due_date,
          ],
        ]);
        $remote_mapping->save();

        if ($job_item->getJob()->isContinuous()) {
          $job_item->active();
        }
      }
      $this->assignMemsourceProviders($project_id, $job);
    }
    catch (TMGMTException $e) {
      $job->rejected('Job has been rejected with following error: @error',
        ['@error' => $e->getMessage()], 'error');
      if (isset($remote_mapping)) {
        $remote_mapping->delete();
      }
    }

    return $job;
  }

  /**
   * Performs a login to Memsource Cloud.
   *
   * @return bool
   *   A success or failure.
   */
  public function loginToMemsource() {
    $config = \Drupal::configFactory()->get('tmgmt_memsource.settings');
    $params = [
      'userName' => $this->translator->getSetting('memsource_user_name'),
      'password' => $this->translator->getSetting('memsource_password'),
    ];
    try {
      $result = $this->request('/api2/v1/auth/login', 'POST', $params);
      if ($result['token']) {
        // Store the token.
        $this->storeToken($result['token']);
        return TRUE;
      }
    }
    catch (TMGMTException $ex) {
      if ($config->get('debug')) {
        \Drupal::logger('tmgmt_memsource')
          ->warning('Unable to log in to Memsource API: ' . $ex->getMessage());
      }
    }
    return FALSE;
  }

  /**
   * Stores a Memsource API token.
   *
   * @param string $token
   *   Token.
   */
  public function storeToken($token) {
    $config = \Drupal::configFactory()->getEditable('tmgmt_memsource.settings');
    $config->set('memsource_token', $token)->save();
  }

  /**
   * Returns a Memsource API token.
   *
   * @return string
   *   Token.
   */
  public function getToken() {
    return \Drupal::configFactory()
      ->get('tmgmt_memsource.settings')
      ->get('memsource_token');
  }

  /**
   * Sends a request to the Memsource API and refreshes the token if necessary.
   *
   * @param string $path
   *   API path.
   * @param string $method
   *   (Optional) HTTP method.
   * @param array $params
   *   (Optional) API params.
   * @param bool $download
   *   (Optional) If true, return the response body as a downloaded content.
   * @param bool $code
   *   (Optional) If true, return only the response HTTP status code.
   * @param string $body
   *   (Optional) An optional request body.
   *
   * @return array|int|null
   *   Result of the API request.
   *
   * @throws \Drupal\tmgmt\TMGMTException
   */
  public function sendApiRequest($path, $method = 'GET', array $params = [], $download = FALSE, $code = FALSE, $body = NULL) {
    $result = NULL;
    $params['token'] = $this->getToken();
    try {
      $result = $this->request($path, $method, $params, $download, $code, $body);
    }
    catch (TMGMTException $ex) {
      if ($ex->getCode() == 401) {
        // Token is invalid, try to re-login.
        $this->loginToMemsource();
        $params['token'] = $this->getToken();
        $result = $this->request($path, $method, $params, $download, $code, $body);
      }
      else {
        throw $ex;
      }
    }
    return $result;
  }

  /**
   * Does a request to Memsource API.
   *
   * @param string $path
   *   Resource path, for example '/api2/v1/auth/login'.
   * @param string $method
   *   (Optional) HTTP method (GET, POST...). By default uses GET method.
   * @param array $params
   *   (Optional) Form parameters to send to Memsource API.
   * @param bool $download
   *   (Optional) If we expect resource to be downloaded. FALSE by default.
   * @param bool $code
   *   (Optional) If we want to return the status code of the call. FALSE by
   *   default.
   * @param string $body
   *   (Optional) Body of the POST request. NULL by
   *   default.
   *
   * @return array|int
   *   Response array or status code.
   *
   * @throws TMGMTException
   */
  public function request($path, $method = 'GET', array $params = [], $download = FALSE, $code = FALSE, $body = NULL) {
    $options = [];

    if (!$this->translator) {
      throw new TMGMTException('There is no Translator entity. Access to the Memsource API is not possible.');
    }

    $config = \Drupal::configFactory()->get('tmgmt_memsource.settings');
    $service_url = $this->translator->getSetting('service_url');

    if (!$service_url) {
      if ($config->get('debug')) {
        \Drupal::logger('tmgmt_memsource')
          ->warning('Attempt to call Memsource API when service_url is not set: ' . $path);
      }
      return [];
    }

    $url = $this->updateServiceUrl($service_url) . $path;

    if ($body) {
      $options['body'] = $body;
      if (isset($params['token'])) {
        $options['query']['token'] = $params['token'];
        unset($params['token']);
      }
      if (!empty($params)) {
        $options['headers'] = $params;
      }
    }
    elseif ($method === 'GET') {
      $options['query'] = $params;
    }
    else {
      if (isset($params['token'])) {
        $options['query']['token'] = $params['token'];
        unset($params['token']);
      }
      $options['json'] = $params;
    }

    $log = [
      '%method' => $method,
      '%url' => $url,
      '%request' => json_encode($options),
    ];

    try {
      $response = $this->client->request($method, $url, $options);
    }
    catch (RequestException $e) {
      if ($config->get('debug')) {
        \Drupal::logger('tmgmt_memsource')->debug(
          "=> REQUEST:\n%method %url\n%request\n\n=> RESPONSE:\n%response\n=> CODE:\n%code",
          $log + [
            '%response' => $e->hasResponse() ? $e->getResponse()->getBody()->getContents() : 'NULL',
            '%code' => $e->getCode(),
          ]
        );
      }

      if (!$e->hasResponse()) {
        if ($code) {
          return $e->getCode();
        }

        throw new TMGMTException('Unable to connect to Memsource API due to following error: @error', ['@error' => $e->getMessage()], $e->getCode());
      }

      $response = $e->getResponse();

      if ($config->get('debug')) {
        \Drupal::logger('tmgmt_memsource')->debug('=> ERROR: ' . "\n" . '%response', [
          '%response' => $e->getResponse()->getBody()->getContents(),
        ]
        );
      }

      if ($code) {
        return $response->getStatusCode();
      }

      throw new TMGMTException('Unable to connect to Memsource API due to following error: @error', ['@error' => $response->getReasonPhrase()], $response->getStatusCode());
    }

    $received_data = $response->getBody()->getContents();

    if ($config->get('debug')) {
      \Drupal::logger('tmgmt_memsource')->debug(
        "=> REQUEST:\n%method %url\n%request\n\n=> RESPONSE:\n%response",
        $log + ['%response' => $received_data]
      );
    }

    if ($code) {
      return $response->getStatusCode();
    }

    if (!in_array($response->getStatusCode(), [200, 201], TRUE)) {
      throw new TMGMTException('Unable to connect to the Memsource API due to following error: @error at @url',
        ['@error' => $response->getStatusCode(), '@url' => $url]);
    }

    if ($download) {
      return $received_data;
    }

    return json_decode($received_data, TRUE);
  }

  /**
   * Modify service URL due to the Memsource API migration (if necessary).
   *
   * 'https://qa.memsource.com/web/api' -> 'https://qa.memsource.com/web'.
   *
   * @param string $url
   *   Service URL.
   *
   * @return string
   *   Updated Memsource URL.
   */
  private function updateServiceUrl($url) {
    $pos = strpos($url, '/web/');

    if ($pos !== FALSE) {
      $url = substr($url, 0, $pos + 4);
      // $this->translator->setSetting('service_url', $url);.
    }

    return $url;
  }

  /**
   * Creates new translation project at Memsource Cloud.
   *
   * @param \Drupal\tmgmt\JobInterface $job
   *   The job.
   *
   * @return string
   *   Memsource project uid.
   */
  public function newTranslationProject(JobInterface $job) {
    $due_date = $job->getSetting('due_date');
    $template_id = $job->getSetting('project_template');

    $params = [
      'name' => $job->label() ?: 'Drupal TMGMT project ' . $job->id(),
      'sourceLang' => $job->getRemoteSourceLanguage(),
      'targetLangs' => [$job->getRemoteTargetLanguage()],
    ];

    if (strlen($due_date) > 0) {
        $params['dateDue'] = $due_date . 'T00:00:00Z';
    }

    if ($template_id === '0' || $template_id === NULL) {
      $result = $this->sendApiRequest('/api2/v1/projects', 'POST', $params);
    }
    else {
      $result = $this->sendApiRequest("/api2/v2/projects/applyTemplate/$template_id", 'POST', $params);
    }

    return $result['uid'];
  }

  /**
   * Assign providers defined in project template.
   *
   * @param string $project_id
   *   Project UID.
   * @param JobInterface $job
   *   The job.
   */
  public function assignMemsourceProviders($project_id, JobInterface $job) {
    $template_id = $job->getSetting('project_template');

    if ($template_id !== NULL && $template_id !== '0') {
      $this->sendApiRequest(
        "/api2/v1/projects/$project_id/applyTemplate/$template_id/assignProviders",
        'POST'
      );
    }
  }

  /**
   * Send the files to Memsource Cloud.
   *
   * @param \Drupal\tmgmt\JobItemInterface $job_item
   *   The Job.
   * @param int $project_id
   *   Memsource project id.
   * @param string $due_date
   *   The date by when the translation is required.
   *
   * @return string
   *   Memsource jobPartId.
   */
  private function sendFiles(JobItemInterface $job_item, $project_id, $due_date) {
    /** @var \Drupal\tmgmt_file\Format\FormatInterface $xliff_converter */
    $xliff_converter = \Drupal::service('plugin.manager.tmgmt_file.format')->createInstance('xlf');

    $job_item_id = $job_item->id();
    $target_language = $job_item->getJob()->getRemoteTargetLanguage();
    $conditions = ['tjiid' => ['value' => $job_item_id]];
    $xliff = $xliff_converter->export($job_item->getJob(), $conditions);
    $name = "JobID_{$job_item->getJob()->id()}_JobItemID_{$job_item_id}_{$job_item->getJob()->getSourceLangcode()}_{$target_language}";

    $job_part_id = $this->createJob($project_id, $target_language, $due_date, $xliff, $name);

    return $job_part_id;
  }

  /**
   * Create a job in Memsource.
   *
   * @param string $project_id
   *   Project ID.
   * @param string $target_language
   *   Target language code.
   * @param string $due_date
   *   Job due date.
   * @param string $xliff
   *   XLIFF file.
   * @param string $name
   *   File name.
   *
   * @return string
   *   Job UID.
   *
   * @throws TMGMTException
   */
  public function createJob($project_id, $target_language, $due_date, $xliff, $name) {
    $params = [
      'targetLangs' => [$target_language],
      'useProjectFileImportSettings' => 'true',
    ];

    if (strlen($due_date) > 0) {
        $params['due'] = $due_date . 'T00:00:00Z';
    }

    $headers = [
      'Content-Disposition' => "filename*=UTF-8''$name.xliff",
      'Memsource' => json_encode($params),
    ];

    $result = $this->sendApiRequest("/api2/v1/projects/$project_id/jobs", 'POST', $headers, FALSE, FALSE, $xliff);

    return $this->getUidOfLatestJob($result['jobs']);
  }

  /**
   * Get UID of the latest returned job (according to workflow steps).
   *
   * @param array $jobs
   *   Jobs.
   *
   * @return string
   *   Job UID.
   */
  private function getUidOfLatestJob($jobs) {
    $latestJob = current($jobs);

    if (count($jobs) > 1) {
      $maxWorkflowLevel = max(array_column($jobs, 'workflowLevel'));

      $filteredJobs = array_filter($jobs, static function ($job) use ($maxWorkflowLevel) {
        return $job['workflowLevel'] == $maxWorkflowLevel;
      });

      if (count($filteredJobs) > 0) {
        $latestJob = current($filteredJobs);
      }
    }

    return $latestJob['uid'];
  }

  /**
   * Fetches translations for job items of a given job.
   *
   * @param \Drupal\tmgmt\JobInterface $job
   *   A job containing job items that translations will be fetched for.
   *
   * @return array
   *   Array containing a containing the number of items translated and the
   *   number that has not been translated yet.
   */
  public function fetchTranslatedFiles(JobInterface $job) {
    $this->setTranslator($job->getTranslator());
    $translated = 0;
    $errors = [];

    try {
      /** @var \Drupal\tmgmt\JobItemInterface $job_item */
      foreach ($job->getItems() as $job_item) {
        $mappings = RemoteMapping::loadByLocalData($job->id(), $job_item->id());
        /** @var \Drupal\tmgmt\Entity\RemoteMapping $mapping */
        foreach ($mappings as $mapping) {
          // Prepare parameters for Job API (to get the job status).
          $job_uid = $mapping->getRemoteIdentifier3();
          $project_id = $mapping->getRemoteIdentifier2();
          $info = [];
          try {
            $info = $this->sendApiRequest("/api2/v1/projects/$project_id/jobs/$job_uid");
          }
          catch (TMGMTException $e) {
            $job->addMessage('Error fetching the job item @job_item. Memsource job @job_uid not found: @error',
              [
                '@job_item' => $job_item->label(),
                '@job_uid' => $job_uid,
                '@error' => $e->getMessage(),
              ], 'error');
            $errors[] = 'Memsource job ' . $job_uid . ' not found, it was probably deleted.';
          }
          if (isset($info['status']) && $this->remoteTranslationCompleted($info['status'])) {
            try {
              $this->addFileDataToJob($job, $info['status'], $project_id, $job_uid);
              $translated++;
            }
            catch (TMGMTException $e) {
              $job->addMessage('Error fetching the job item @job_item: @error', ['@job_item' => $job_item->label(), '@error' => $e->getMessage()], 'error');
              continue;
            }
          }
        }
      }
    }
    catch (TMGMTException $e) {
      \Drupal::logger('tmgmt_memsource')
        ->error('Could not pull translation resources: @error', ['@error' => $e->getMessage()]);
    }

    return [
      'translated' => $translated,
      'untranslated' => count($job->getItems()) - $translated,
      'errors' => $errors,
    ];
  }

  /**
   * Retrieve all the updates for all the job items in a translator.
   *
   * @param \Drupal\tmgmt\JobItemInterface $job_item
   *   The job item to get the translation.
   *
   * @return int
   *   The number of updated job items.
   */
  public function pullRemoteTranslation(JobItemInterface $job_item) {
    $job = $job_item->getJob();
    $this->setTranslator($job->getTranslator());
    $remotes = RemoteMapping::loadByLocalData($job->id(), $job_item->id());
    $remote = reset($remotes);
    $job_uid = $remote->getRemoteIdentifier3();
    $project_id = $remote->getRemoteIdentifier2();
    $info = $this->sendApiRequest("/api2/v1/projects/$project_id/jobs/$job_uid");
    if (isset($info['status']) && $this->remoteTranslationCompleted($info['status'])) {
      try {
        $this->addFileDataToJob($job, $info['status'], $project_id, $job_uid);
        return 1;
      }
      catch (TMGMTException $e) {
        $job->addMessage('Error fetching the job item: @job_item.', [
          '@job_item' => $remote->getJobItem()->label(),
        ], 'error');
      }
    }
    return 0;
  }

  /**
   * Retrieve the data of a file in a state.
   *
   * @param \Drupal\tmgmt\JobInterface $job
   *   The Job to which will be added the data.
   * @param string $state
   *   The state of the file.
   * @param int $project_id
   *   The project ID.
   * @param string $job_uid
   *   The file ID.
   *
   * @throws \Drupal\tmgmt\TMGMTException
   */
  public function addFileDataToJob(JobInterface $job, $state, $project_id, $job_uid) {
    $data = $this->sendApiRequest("/api2/v1/projects/$project_id/jobs/$job_uid/targetFile", 'GET', [], TRUE);
    $file_data = $this->parseTranslationData($data);
    if ($this->remoteTranslationCompleted($state)) {
      $status = TMGMT_DATA_ITEM_STATE_TRANSLATED;
    }
    else {
      $status = TMGMT_DATA_ITEM_STATE_PRELIMINARY;
    }
    $job->addTranslatedData($file_data, [], $status);
    $mappings = RemoteMapping::loadByRemoteIdentifier('tmgmt_memsource', $project_id, $job_uid);
    $mapping = reset($mappings);
    $mapping->removeRemoteData('TmsState');
    $mapping->addRemoteData('TmsState', $state);
    $mapping->save();
  }

  /**
   * Parses received translation from Memsource and returns unflatted data.
   *
   * @param string $data
   *   Xliff data, received from Memsource Cloud.
   *
   * @return array
   *   Unflatted data.
   */
  protected function parseTranslationData($data) {
    /** @var \Drupal\tmgmt_file\Format\FormatInterface $xliff_converter */
    $xliff_converter = \Drupal::service('plugin.manager.tmgmt_file.format')->createInstance('xlf');
    // Import given data using XLIFF converter.
    // Specify that passed content is not a file.
    return $xliff_converter->import($data, FALSE);
  }

  /**
   * Check that given status represents completed state.
   *
   * @param string $status
   *   Given status.
   *
   * @return bool
   *   Return true if status is conmleted.
   */
  public function remoteTranslationCompleted($status) {
    return in_array($status, ['COMPLETED_BY_LINGUIST', 'COMPLETED', 'DELIVERED']);
  }

}
